"""
KIRA interface to send UDP packets to a Keene Electronics IR-IP bridge.
"""
# pylint: disable=import-error
import socket
import sys
import time

from . import utils

from .constants import (
    DEFAULT_PORT
)

class KiraModule(object):
    """
    Construct and send IR-IP packets.
    """

    def __init__(self, host, port=DEFAULT_PORT):
        """Construct a KIRA interface object."""
        self.host = host
        self.port = port
        self.codeMap = {}

    def registerCode(self, codeName, code, codeType=None):
        rawCode = utils.code2kira(code, codeType=codeType)
        if rawCode:
            self.codeMap[codeName] = rawCode

    def sendCode(self, codeName, repeat=1, delay=0.05):
        code = self.codeMap.get(codeName)
        if code:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect((self.host, self.port))
            if sys.version_info[0] == 2:
                code = "%s\n" % code
            else:
                code = bytes("%s\n" % code, "ascii")
            for i in range(repeat):
                sock.send(code)
                time.sleep(delay)
            sock.close()



"""
KIRA interface to receive UDP packets from a Keene Electronics IR-IP bridge.
"""
# pylint: disable=import-error
import os
import select
import socket
import sys
import threading
import time

from . import utils

from .constants import (
    DEFAULT_HOST,
    DEFAULT_PORT
)


class KiraReceiver(threading.Thread):
    """
    Open a UDP socket to monitor for incoming IR-IP packets.
    """

    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT):
        """Construct a KIRA interface object."""
        threading.Thread.__init__(self)
        self.stopped = threading.Event()
        self._callbacks = []
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.host = host
        self.port = port
        self._socket.bind((host, port))
        self.codeMap = {}
        self._state = 'idle'

    def registerCode(self, codeName, code, codeType=None):
        code = utils.mangleCode(code, codeType=codeType)
        if code:
            self.codeMap[code] = codeName

    def registerCallback(self, callback):
        self._callbacks.append(callback)

    def run(self):
        """Main loop of KIRA thread."""
        while not self.stopped.isSet():
            try:
                # if the current state is idle, just block and wait forever
                # if the current state is any other state, then a timeout of 200ms should
                # be reasonable in all cases.
                timeout = (self._state != 'idle') and 0.2 or None
                rdlist, _, _ = select.select([self._socket.fileno()], [], [], timeout)
                if not rdlist:
                    if self._state != 'idle':
                        self._state = 'idle'
                    continue
                data = self._socket.recv(1024)
                if not data:
                    # check if the socket is still valid
                    try:
                        os.fstat(recv._socket.fileno())
                    except socket.error:
                        break
                    continue
                code = utils.mangleIR(data, ignore_errors=True)
                codeName = self.codeMap.get(code)
                # some manufacturers repeat their IR codes several times in rapid
                # succession. by tracking the last code, we can eliminate redundant
                # state changes
                if codeName and (self._state != codeName):
                    self._state = codeName
                    for callback in self._callbacks:
                        callback(codeName)
            except:
                time.sleep(0.1)

    def stop(self):
        self.stopped.set()
        # force receiver thread to wake from select
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect((self.host, self.port))
        msg = "stop"
        datagram = sys.version_info[0] == 2 and bytes(msg) or bytes(msg, "utf-8")
        sock.send(datagram)
        sock.close()
        self.join()
        self._socket.close()



import binascii
import re

# this mangled "shorthand" output basically throws out exact timings while
# keeping the relative length of each time period. Good for comparing
# received codes against known values
def mangleIR(data, ignore_errors=False):
    """Mangle a raw Kira data packet into shorthand"""
    try:
        # Packet mangling algorithm inspired by Rex Becket's kirarx vera plugin
        # Determine a median value for the timing packets and categorize each
        # timing as longer or shorter than that. This will always work for signals
        # that use pulse width modulation (since varying by long-short is basically
        # the definition of what PWM is). By lucky coincidence this also works with
        # the RC-5/RC-6 encodings used by Phillips (manchester encoding)
        # because time variations of opposite-phase/same-phase are either N or 2*N
        if isinstance(data, bytes):
            data = data.decode('ascii')
        data = data.strip()
        times = [int(x, 16) for x in data.split()[2:]]
        minTime = min(times[2:-1])
        maxTime = max(times[2:-1])
        margin = (maxTime - minTime) / 2 + minTime
        return ''.join([(x < margin and 'S' or 'L') for x in times])
    except:
        # Probably a mangled packet.
        if not ignore_errors:
            raise

# pronto codes provide basically the same information as a Kira code
# just in a slightly different form. (Kira represents timings in uS
# while pronto uses multiples of the base clock cycle.)
# Thus they can be used for transmission
def pronto2kira(data):
    """Convert a pronto code to a discrete (single button press) Kira code"""
    octets = [int(x, 16) for x in data.split()]
    preamble = octets[:4]
    convert = lambda x: 1000.0 / (x * 0.241246)
    freq = convert(preamble[1])
    period = 1000000.0 / (freq * 1000.0)
    dataLen = preamble[2]
    res = "K %02X%02X " %(freq, dataLen)
    res += " ".join(["%0.4X" % min(0x2000, (period * x)) for x in octets[4: 4+(2*dataLen)]])
    return res

def mangleNec(code, freq=40):
    """Convert NEC code to shorthand notation"""
    # base time is 550 microseconds
    # unit of burst time
    # lead in pattern:   214d 10b3
    # "1" burst pattern: 0226 0960
    # "0" burst pattern: 0226 0258
    # lead out pattern:  0226 2000
    # there's large disagreement between devices as to a common preamble
    # or the "long" off period for the representation of a binary 1
    # thus we can't construct a code suitable for transmission
    # without more information--but it's good enough for creating
    # a shorthand representaiton for use with recv
    timings = []
    for octet in binascii.unhexlify(code.replace(" ", "")):
        burst = lambda x: x and "0226 06AD" or "0226 0258"
        for bit in reversed("%08d" % int(bin(ord(octet))[2:])):
            bit = int(bit)
            timings.append(burst(bit))
    return mangleIR("K %0X22 214d 10b3 " % freq + " ".join(timings) + " 0226 2000")

def inferCodeType(data):
    # a series of L/S chars
    if re.match('^[LS]+$', data):
        return 'shorthand'
    # "K " followed by groups of 4 hex chars
    if re.match("^K ([0-9a-fA-F]{4} )*[0-9a-fA-F]{4}$", data):
        return 'kira'
    # 2 groups of 4 hex chars
    if re.match("^[0-9a-fA-F]{4} [0-9a-fA-F]{4}$", data):
        return 'nec'
    # multiple groups of 4 hex chars
    if re.match("^([0-9a-fA-F]{4} )*[0-9a-fA-F]{4}$", data):
        return 'pronto'

# convert code into a raw kira code that can be transmitted
def code2kira(code, codeType=None):
    if codeType is None:
        codeType = inferCodeType(code)
    if codeType == "kira":
        return code.strip()
    if codeType == "pronto":
        return pronto2kira(code)

# convert code to a form ready for comparison
def mangleCode(code, codeType=None):
    if codeType is None:
        codeType = inferCodeType(code)
    if codeType == "shorthand":
        return code.strip()
    if codeType == "kira":
        return mangleIR(code)
    if codeType == "pronto":
        return mangleIR(pronto2kira(code))
    if codeType == "nec":
        return mangleNec(code)
